from importlib.util import find_spec

from daml._internal.metrics.ber import BER
from daml._internal.metrics.divergence import Divergence
from daml._internal.metrics.uap import UAP
from daml.metrics import drift

__all__ = ["BER", "Divergence", "UAP", "drift"]

if find_spec("tensorflow") is not None and find_spec("tensorflow_probability") is not None:  # pragma: no cover
    from . import outlier

    __all__ += ["outlier"]

if find_spec("torch") is not None:  # pragma: no cover
    from daml._internal.metrics.sufficiency import Sufficiency

    __all__ += ["Sufficiency"]

del find_spec



from importlib.util import find_spec

from daml._internal.metrics.drift.base import LastSeenUpdate, ReservoirSamplingUpdate
from daml._internal.metrics.drift.cvm import CVMDrift
from daml._internal.metrics.drift.ks import KSDrift

__all__ = ["CVMDrift", "KSDrift", "LastSeenUpdate", "ReservoirSamplingUpdate"]

if find_spec("torch") is not None:  # pragma: no cover
    from daml._internal.metrics.drift.mmd import MMDDrift
    from daml._internal.metrics.drift.torch import GaussianRBF, preprocess_drift
    from daml._internal.metrics.drift.uncertainty import UncertaintyDrift

    __all__ += ["MMDDrift", "GaussianRBF", "UncertaintyDrift", "preprocess_drift"]

del find_spec



from daml._internal.metrics.outlier.ae import AEOutlier
from daml._internal.metrics.outlier.aegmm import AEGMMOutlier
from daml._internal.metrics.outlier.base import OutlierScore
from daml._internal.metrics.outlier.llr import LLROutlier
from daml._internal.metrics.outlier.vae import VAEOutlier
from daml._internal.metrics.outlier.vaegmm import VAEGMMOutlier

__all__ = [
    "AEOutlier",
    "AEGMMOutlier",
    "LLROutlier",
    "OutlierScore",
    "VAEOutlier",
    "VAEGMMOutlier",
]



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from typing import Callable

import numpy as np
from keras.api._v2 import keras

from daml._internal.metrics.outlier.base import BaseOutlier, OutlierScore
from daml._internal.models.tensorflow.autoencoder import VAE
from daml._internal.models.tensorflow.losses import Elbo
from daml._internal.models.tensorflow.utils import predict_batch


class VAEOutlier(BaseOutlier):
    def __init__(self, model: VAE, samples: int = 10) -> None:
        """
        VAE based outlier detector.

        Parameters
        ----------
        model : VAE
            A VAE model.
        samples : int, default 10
            Number of samples sampled to evaluate each instance.
        """
        super().__init__(model)
        self.samples = samples

    def fit(
        self,
        x_ref: np.ndarray,
        threshold_perc: float = 100.0,
        loss_fn: Callable = Elbo(0.05),
        optimizer: keras.optimizers.Optimizer = keras.optimizers.Adam,
        epochs: int = 20,
        batch_size: int = 64,
        verbose: bool = True,
    ) -> None:
        """
        Train the VAE model.

        Parameters
        ----------
        x_ref : np.ndarray
            Training batch.
        threshold_perc : float, default 100.0
            Percentage of reference data that is normal.
        loss_fn : Callable, default Elbo(0.05)
            Loss function used for training.
        optimizer : keras.optimizers.Optimizer, default keras.optimizers.Adam
            Optimizer used for training.
        epochs : int, default 20
            Number of training epochs.
        batch_size : int, default 64
            Batch size used for training.
        verbose : bool, default True
            Whether to print training progress.
        """
        super().fit(x_ref, threshold_perc, loss_fn, optimizer, epochs, batch_size, verbose)

    def score(self, X: np.ndarray, batch_size: int = int(1e10)) -> OutlierScore:
        self._validate(X)

        # sample reconstructed instances
        X_samples = np.repeat(X, self.samples, axis=0)
        X_recon = predict_batch(X_samples, model=self.model, batch_size=batch_size)

        # compute feature scores
        fscore = np.power(X_samples - X_recon, 2)
        fscore = fscore.reshape((-1, self.samples) + X_samples.shape[1:])
        fscore = np.mean(fscore, axis=1)

        # compute instance scores
        fscore_flat = fscore.reshape(fscore.shape[0], -1).copy()
        n_score_features = int(np.ceil(fscore_flat.shape[1]))
        sorted_fscore = np.sort(fscore_flat, axis=1)
        sorted_fscore_perc = sorted_fscore[:, -n_score_features:]
        iscore = np.mean(sorted_fscore_perc, axis=1)

        return OutlierScore(iscore, fscore)



from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union, cast

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
from matplotlib.figure import Figure
from scipy.optimize import basinhopping
from torch.utils.data import Dataset

from daml._internal.metrics.base import EvaluateMixin

STEPS_KEY = "_STEPS_"


def f_out(n_i: np.ndarray, x: np.ndarray) -> np.ndarray:
    """
    Calculates the line of best fit based on its free parameters

    Parameters
    ----------
    n_i : np.ndarray
        Array of sample sizes
    x : np.ndarray
        Array of inverse power curve coefficients

    Returns
    -------
    np.ndarray
        Data points for the line of best fit
    """
    return x[0] * n_i ** (-x[1]) + x[2]


def f_inv_out(y_i: np.ndarray, x: np.ndarray) -> np.ndarray:
    """
    Inverse function for f_out()

    Parameters
    ----------
    y_i : np.ndarray
        Data points for the line of best fit
    x : np.ndarray
        Array of inverse power curve coefficients

    Returns
    -------
    np.ndarray
        Array of sample sizes
    """
    n_i = ((y_i - x[2]) / x[0]) ** (-1 / x[1])
    return n_i


def calc_params(p_i: np.ndarray, n_i: np.ndarray, niter: int) -> np.ndarray:
    """
    Retrieves the inverse power curve coefficients for the line of best fit.
    Global minimization is done via basin hopping. More info on this algorithm
    can be found here: https://arxiv.org/abs/cond-mat/9803344 .

    Parameters
    ----------
    p_i : np.ndarray
        Array of corresponding losses
    n_i : np.ndarray
        Array of sample sizes
    niter : int
        Number of iterations to perform in the basin-hopping
        numerical process to curve-fit p_i

    Returns
    -------
    np.ndarray
        Array of parameters to recreate line of best fit
    """

    def f(x):
        return np.sum(np.square(p_i - x[0] * n_i ** (-x[1]) - x[2]))

    res = basinhopping(f, np.array([0.5, 0.5, 0.1]), niter=niter)
    return res.x


def reset_parameters(model: nn.Module):
    """
    Re-initializes each layer in the model using
    the layer's defined weight_init function
    """

    @torch.no_grad()
    def weight_reset(m: nn.Module):
        # Check if the current module has reset_parameters
        reset_parameters = getattr(m, "reset_parameters", None)
        if callable(reset_parameters):
            m.reset_parameters()  # type: ignore

    # Applies fn recursively to every submodule see:
    # https://pytorch.org/docs/stable/generated/torch.nn.Module.html
    return model.apply(fn=weight_reset)


def validate_dataset_len(dataset: Dataset) -> int:
    if not hasattr(dataset, "__len__"):
        raise TypeError("Must provide a dataset with a length attribute")
    length: int = dataset.__len__()  # type: ignore
    if length <= 0:
        raise ValueError("Dataset length must be greater than 0")
    return length


def validate_output(data: Dict[str, np.ndarray]):
    """Ensure the sufficiency data used is not malformed"""
    if STEPS_KEY not in data:
        raise KeyError(f"{STEPS_KEY} is a required key for Sufficiency output.")
    c = len(data[STEPS_KEY])
    for m, v in data.items():
        if m == STEPS_KEY:
            continue
        if c != len(v):
            raise ValueError("f{m} does not contain the expected number ({c}) of data points.")


def project_steps(
    measure: np.ndarray,
    steps: np.ndarray,
    projection: np.ndarray,
    niter: int = 1000,
) -> Tuple[np.ndarray, np.ndarray]:
    """Projects the measures for each value of X

    Parameters
    ----------
    measure : np.ndarray
        Measures from which to extrapolate projection
    steps : np.ndarray
        Steps of the taken measures
    projection : np.ndarray
        Steps to extrapolate
    niter : int, default 1000
        Number of iterations to perform in the basin-hopping
        numerical process to curve-fit measure

    Returns
    -------
    np.ndarray
        Extrapolated measure values at each projection step
    np.ndarray
        length-3 array of the parameters for the fit curve.

    """
    params = calc_params(p_i=(1 - measure), n_i=steps, niter=niter)
    projected_steps = 1 - f_out(projection, params)
    return projected_steps, params


def inv_project_steps(
    measure: np.ndarray,
    steps: np.ndarray,
    accuracies: np.ndarray,
    params: np.ndarray = np.zeros(0),
) -> np.ndarray:
    """Inverse function for project_steps()

    Parameters
    ----------
    measure : np.ndarray
        Measures from which to extrapolate projection
    steps : np.ndarray
        Steps of the taken measures
    accuracies : np.ndarray
        Desired accuracy values
    params : np.ndarray, default np.zeros(0)
        length-3 array of the parameters for the sufficiency curve to study.
        If not provided, we curve-fit the parameters at runtime. Curve-fitting
        is very slow, and we recommend pre-computing the parameters.

    Returns
    -------
    np.ndarray
        Array of sample sizes
    """
    if len(params) == 0:
        params = calc_params(p_i=(1 - measure), n_i=steps, niter=1000)
    return f_inv_out(1 - np.array(accuracies), params)


def plot_measure(
    name: str,
    steps: np.ndarray,
    measure: np.ndarray,
    projection: np.ndarray,
) -> Figure:
    fig = plt.figure()
    fig = cast(Figure, fig)
    fig.tight_layout()

    ax = fig.add_subplot(111)

    ax.set_title(f"{name} Sufficiency")
    ax.set_ylabel(f"{name}")
    ax.set_xlabel("Steps")

    # Plot measure over each step
    ax.scatter(steps, measure, label=f"Model Results ({name})", s=15, c="black")

    # Plot extrapolation
    ax.plot(
        projection,
        project_steps(measure, steps, projection)[0],
        linestyle="dashed",
        label=f"Potential Model Results ({name})",
    )

    ax.legend()
    return fig


class Sufficiency(EvaluateMixin):
    """
    Project dataset sufficiency using given a model and evaluation criteria

    Parameters
    ----------
    model : nn.Module
        Model that will be trained for each subset of data
    train_ds : Dataset
        Full training data that will be split for each run
    test_ds : Dataset
        Data that will be used for every run's evaluation
    train_fn : Callable[[nn.Module, Dataset, Sequence[int]], None]
        Function which takes a model (torch.nn.Module), a dataset
        (torch.utils.data.Dataset), indices to train on and executes model
        training against the data.
    eval_fn : Callable[[nn.Module, Dataset], Dict[str, float]]
        Function which takes a model (torch.nn.Module), a dataset
        (torch.utils.data.Dataset) and returns a dictionary of metric
        values (Dict[str, float]) which is used to assess model performance
        given the model and data.
    runs : int, default 1
        Number of models to run over all subsets
    substeps : int, default 5
        Total number of dataset partitions that each model will train on
    train_kwargs : Dict[str, Any] | None, default None
        Additional arguments required for custom training function
    eval_kwargs : Dict[str, Any] | None, default None
        Additional arguments required for custom evaluation function
    """

    def __init__(
        self,
        model: nn.Module,
        train_ds: Dataset,
        test_ds: Dataset,
        train_fn: Callable[[nn.Module, Dataset, Sequence[int]], None],
        eval_fn: Callable[[nn.Module, Dataset], Union[Dict[str, float], Dict[str, np.ndarray]]],
        runs: int = 1,
        substeps: int = 5,
        train_kwargs: Optional[Dict[str, Any]] = None,
        eval_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self.model = model
        self.train_ds = train_ds
        self.test_ds = test_ds
        self.train_fn = train_fn
        self.eval_fn = eval_fn
        self.runs = runs
        self.substeps = substeps
        self.train_kwargs = train_kwargs
        self.eval_kwargs = eval_kwargs

    @property
    def train_ds(self):
        return self._train_ds

    @train_ds.setter
    def train_ds(self, value: Dataset):
        self._train_ds = value
        self._length = validate_dataset_len(value)

    @property
    def test_ds(self):
        return self._test_ds

    @test_ds.setter
    def test_ds(self, value: Dataset):
        validate_dataset_len(value)
        self._test_ds = value

    @property
    def train_fn(self) -> Callable[[nn.Module, Dataset, Sequence[int]], None]:
        return self._train_fn

    @train_fn.setter
    def train_fn(self, value: Callable[[nn.Module, Dataset, Sequence[int]], None]):
        if not callable(value):
            raise TypeError("Must provide a callable for train_fn.")
        self._train_fn = value

    @property
    def eval_fn(
        self,
    ) -> Callable[[nn.Module, Dataset], Union[Dict[str, float], Dict[str, np.ndarray]]]:
        return self._eval_fn

    @eval_fn.setter
    def eval_fn(
        self,
        value: Callable[[nn.Module, Dataset], Union[Dict[str, float], Dict[str, np.ndarray]]],
    ):
        if not callable(value):
            raise TypeError("Must provide a callable for eval_fn.")
        self._eval_fn = value

    @property
    def train_kwargs(self) -> Dict[str, Any]:
        return self._train_kwargs

    @train_kwargs.setter
    def train_kwargs(self, value: Optional[Dict[str, Any]]):
        self._train_kwargs = {} if value is None else value

    @property
    def eval_kwargs(self) -> Dict[str, Any]:
        return self._eval_kwargs

    @eval_kwargs.setter
    def eval_kwargs(self, value: Optional[Dict[str, Any]]):
        self._eval_kwargs = {} if value is None else value

    def evaluate(self, eval_at: Optional[np.ndarray] = None) -> Dict[str, np.ndarray]:
        """
        Creates data indices, trains models, and returns plotting data

        Inputs
        ------
        eval_at : Optional[np.ndarray]
            Specify this to collect accuracies over a specific set of dataset lengths,
            rather than letting Sufficiency internally create the lengths
            to evaluate at.

        Returns
        -------
        Dict[str, np.ndarray]
            Dictionary containing the average of each measure per substep
        """
        if eval_at is not None:
            ranges = eval_at
        else:
            geomshape = (
                0.01 * self._length,
                self._length,
                self.substeps,
            )  # Start, Stop, Num steps
            ranges = np.geomspace(*geomshape).astype(np.int64)
        substeps = len(ranges)
        metric_outputs = {}

        # Run each model over all indices
        for _ in range(self.runs):
            # Create a randomized set of indices to use
            indices = np.random.randint(0, self._length, size=self._length)
            # Reset the network weights to "create" an untrained model
            model = reset_parameters(self.model)
            # Run the model with each substep of data
            for iteration, substep in enumerate(ranges):
                # train on subset of train data
                self.train_fn(
                    model,
                    self.train_ds,
                    indices[:substep].tolist(),
                    **self.train_kwargs,
                )

                # evaluate on test data
                output = self.eval_fn(
                    model,
                    self.test_ds,
                    **self.eval_kwargs,
                )

                # Keep track of each measures values
                for name, value in output.items():
                    if name == STEPS_KEY:
                        raise KeyError(f"Cannot use '{STEPS_KEY}' as a metric name.")

                    if name not in metric_outputs:
                        shape = (substeps, len(value)) if isinstance(value, np.ndarray) else substeps
                        metric_outputs[name] = np.zeros(shape)

                    # Sum result into current substep iteration to be averaged later
                    metric_outputs[name][iteration] += value

        output = {STEPS_KEY: ranges}
        # The mean for each measure must be calculated before being returned
        output.update({name: value / self.runs for name, value in metric_outputs.items()})

        return output

    @classmethod
    def project(
        cls,
        data: Dict[str, np.ndarray],
        projection: Union[int, Sequence[int], np.ndarray],
        niter: int = 1000,
    ) -> Dict[str, np.ndarray]:
        """Projects the measures for each value of X

        Parameters
        ----------
        data : Dict[str, np.ndarray]
            Dataclass containing the average of each measure per substep
        steps : Union[int, np.ndarray]
            Step or steps to project
        niter : int, default 1000
            Number of iterations to perform in the basin-hopping
            numerical process to curve-fit data

        Raises
        ------
        KeyError
            If STEPS_KEY or measure is not a valid key
        ValueError
            If the length of data points in the measures do not match
            If the steps are not int, Sequence[int] or an ndarray
        """
        validate_output(data)
        projection = [projection] if isinstance(projection, int) else projection
        projection = np.array(projection) if isinstance(projection, Sequence) else projection
        if not isinstance(projection, np.ndarray):
            raise ValueError("'steps' must be an int, Sequence[int] or ndarray")

        output = {}
        output[STEPS_KEY] = projection
        params_cache = []
        for name, measure in data.items():
            if name == STEPS_KEY:
                continue

            if measure.ndim > 1:
                result = []
                for i in range(measure.shape[1]):
                    projected, params = project_steps(measure[:, i], data[STEPS_KEY], projection, niter)
                    params_cache.append(params)
                    result.append(projected)
                output[name] = np.array(result).T
            else:
                output[name], params = project_steps(measure, data[STEPS_KEY], projection, niter)
                params_cache.append(params)
        return output

    @classmethod
    def plot(cls, data: Dict[str, np.ndarray], class_names: Optional[Sequence[str]] = None) -> List[Figure]:
        """Plotting function for data sufficiency tasks

        Parameters
        ----------
        data : Dict[str, np.ndarray]
            Dataclass containing the average of each measure per substep

        Returns
        -------
        List[plt.Figure]
            List of Figures for each measure

        Raises
        ------
        KeyError
            If STEPS_KEY or measure is not a valid key
        ValueError
            If the length of data points in the measures do not match
        """
        validate_output(data)

        # X, y data
        steps = data[STEPS_KEY]

        # Extrapolation parameters
        last_X = steps[-1]
        geomshape = (0.01 * last_X, last_X * 4, len(steps))
        extrapolated = np.geomspace(*geomshape).astype(np.int64)

        # Stores all plots
        plots = []

        # Create a plot for each measure on one figure
        for name, measure in data.items():
            if name == STEPS_KEY:
                continue

            if len(measure.shape) > 1:
                if class_names is not None and measure.shape[1] != len(class_names):
                    raise IndexError("Class name count does not align with measures")
                for i in range(measure.shape[1]):
                    class_name = str(i) if class_names is None else class_names[i]
                    fig = plot_measure(
                        name + "_" + class_name,
                        steps,
                        measure[:, i],
                        extrapolated,
                    )
                    plots.append(fig)

            else:
                fig = plot_measure(name, steps, measure, extrapolated)
                plots.append(fig)

        return plots

    @classmethod
    def inv_project(
        cls, targets: np.ndarray, data: Dict[str, np.ndarray], params_cache: np.ndarray = np.zeros((1, 0))
    ) -> np.ndarray:
        """
        Calculate he number of training samples needed to achieve the target model
        metric values.

        Parameters
        ----------
        targets : np.ndarray
            List of the target metric scores (from 0.0 to 1.0) that we want to achieve.

        data : Dict[str, np.ndarray]
            Dataclass containing the average of each measure per substep

        params_cache : np.ndarray, default np.zeros((1,0))
            1 x 3 List of cached parameters for the sufficiency curve. The parameters
            can be precomputed by using the project() function.
            TODO: The first axis should represent the number of data columns. It's
            currently hardcoded to assume only one column.


        Returns
        -------
        np.ndarray(np.int64)
            List of the number of training samples needed to achieve each
            corresponding entry in targets
        """

        validate_output(data)

        

        # Iterate through the elements of the data dictionary until
        # we reach the array of measures, which are then used to predict
        # the number of needed training samples
        for name, measure in data.items():
            if name == STEPS_KEY:
                continue

            # Select the cached parameters associated with the current column of measure
            # TODO: We currently assume measure only has one axis
            params = params_cache[0]

            # X, y data
            steps = data[STEPS_KEY]

            num_samples_needed = inv_project_steps(measure, steps, targets, params)
            return np.array(np.ceil(num_samples_needed), dtype=np.int64)

        # TODO: Is this a reasonable error response
        return np.array([np.int64(-1)])



from dataclasses import dataclass
from typing import Generic, NamedTuple, Optional, TypeVar, Union

import numpy as np
from scipy.signal import convolve2d
from scipy.stats import entropy, kurtosis, skew

EDGE_KERNEL = np.array([[-1, -1, -1], [-1, 8, -1], [-1, -1, -1]], dtype=np.int8)
BIT_DEPTH = (1, 8, 12, 16, 32)

TStatValue = TypeVar("TStatValue", np.number, np.ndarray)


class BitDepth(NamedTuple):
    depth: int
    pmin: Union[float, int]
    pmax: Union[float, int]


def get_bitdepth(image: np.ndarray) -> BitDepth:
    """
    Approximates the bit depth of the image using the
    min and max pixel values.
    """
    pmin, pmax = np.min(image), np.max(image)
    if pmin < 0:
        return BitDepth(0, pmin, pmax)
    else:
        depth = ([x for x in BIT_DEPTH if 2**x > pmax] or [max(BIT_DEPTH)])[0]
        return BitDepth(depth, 0, 2**depth - 1)


def rescale(image: np.ndarray, depth: int = 1) -> np.ndarray:
    """
    Rescales the image using the bit depth provided.
    """
    bitdepth = get_bitdepth(image)
    if bitdepth.depth == depth:
        return image
    else:
        normalized = (image + bitdepth.pmin) / (bitdepth.pmax - bitdepth.pmin)
        return normalized * (2**depth - 1)


def normalize_image_shape(image: np.ndarray) -> np.ndarray:
    """
    Normalizes the image shape into (C,H,W).
    """
    ndim = image.ndim
    if ndim == 2:
        return np.expand_dims(image, axis=0)
    elif ndim == 3:
        return image
    elif ndim > 3:
        # Slice all but the last 3 dimensions
        return image[(0,) * (ndim - 3)]
    else:
        raise ValueError("Images must have 2 or more dimensions.")


def edge_filter(image: np.ndarray, offset: float = 0.5) -> np.ndarray:
    """
    Returns the image filtered using a 3x3 edge detection kernel:
    [[ -1, -1, -1 ],
     [ -1,  8, -1 ],
     [ -1, -1, -1 ]]
    """
    edges = convolve2d(image, EDGE_KERNEL, mode="same", boundary="symm") + offset
    np.clip(edges, 0, 255, edges)
    return edges


@dataclass
class ImageStats(Generic[TStatValue]):
    """
    Dataclass containing annotations for all of the supported
    metrics for image statistics.
    """

    height: TStatValue
    width: TStatValue
    size: TStatValue
    aspect_ratio: TStatValue
    depth: TStatValue
    channels: TStatValue
    missing: TStatValue
    brightness: TStatValue
    blurriness: TStatValue
    mean: TStatValue
    zero: TStatValue
    var: TStatValue
    skew: TStatValue
    kurtosis: TStatValue
    percentiles: np.ndarray
    histogram: np.ndarray
    entropy: TStatValue


@dataclass
class ChannelStats:
    """
    Dataclass containing annotations for all of the supported
    metrics for channel specific statistics.
    """

    ch_mean: np.ndarray
    ch_var: np.ndarray
    ch_skew: np.ndarray
    ch_kurtosis: np.ndarray
    ch_percentiles: np.ndarray
    ch_histogram: np.ndarray


def get_image_stats(image: np.ndarray) -> ImageStats:
    image = normalize_image_shape(image)
    width = np.int32(image.shape[-1])
    height = np.int32(image.shape[-2])
    scaled = rescale(image)
    histogram = np.histogram(scaled, bins=256, range=(0, 1))[0]

    return ImageStats(
        height=height,
        width=width,
        size=width * height,
        aspect_ratio=width / height,
        depth=np.int32(get_bitdepth(image).depth),
        channels=np.int32(image.shape[-3]),
        missing=np.sum(np.isnan(image)),
        brightness=np.mean(scaled),  # TODO: replace with better calculation
        blurriness=np.std(edge_filter(np.mean(image, axis=0))),
        mean=np.mean(scaled),
        zero=np.int32(np.count_nonzero(image == 0)),
        var=np.var(scaled),
        skew=np.float32(skew(scaled.ravel())),
        kurtosis=np.float32(kurtosis(scaled.ravel())),
        percentiles=np.percentile(scaled, q=[0, 25, 50, 75, 100]),
        histogram=histogram,
        entropy=np.float32(entropy(histogram)),
    )


def get_channel_stats(image: np.ndarray) -> ChannelStats:
    image = normalize_image_shape(image)
    scaled = rescale(image)
    flattened = scaled.reshape(image.shape[0], -1)

    return ChannelStats(
        ch_mean=np.mean(scaled, axis=(1, 2)),
        ch_var=np.var(scaled, axis=(1, 2)),
        ch_skew=skew(flattened, axis=1),
        ch_kurtosis=kurtosis(flattened, axis=1),
        ch_percentiles=np.percentile(scaled, q=[0, 25, 50, 75, 100], axis=(1, 2)).T,
        ch_histogram=np.apply_along_axis(lambda x: np.histogram(x, bins=256, range=(0, 1))[0], 1, flattened),
    )


# Pulls in a dataset and then gets the individual image stats
# then runs group stats
# Class to encapsulate dataset statistics calculations
class DatasetStats(ImageStats[np.ndarray], ChannelStats):
    def __init__(
        self,
        images,
        labels: Optional[np.ndarray] = None,
        boxes: Optional[np.ndarray] = None,
    ) -> None:
        # Initialization of DatasetStats with datasets images, optional labels, and optional bounding boxes.
        # self.images = images
        # self.labels = labels
        # self.boxes = boxes
        self.length = len(images)
        channel_map = []
        channel_stats = {}

        # Iterate through images
        for i, image in enumerate(images):
            # stats = SingleImageStats(image)
            img_stats = get_image_stats(image)

            # Aggregate the image statistics
            for stat in ImageStats.__annotations__:
                image_stat = getattr(img_stats, stat)
                aggregated_stat = getattr(self, stat, None)
                if aggregated_stat is None:
                    shape = () if np.isscalar(image_stat) else image_stat.shape
                    aggregated_stat = np.empty((self.length,) + shape)
                    setattr(self, stat, aggregated_stat)
                aggregated_stat[i] = image_stat

            # Build the image and channel mapping for each channel
            c = int(img_stats.channels)
            channel_map.append(np.linspace((i, c, 0), (i, c, c - 1), c, dtype=np.uint32))

            # Aggregate the channel statistics into channel_stats dictionary
            ch_stats = get_channel_stats(image)
            for stat in ChannelStats.__annotations__:
                channel_stat = getattr(ch_stats, stat)
                channel_stats.setdefault(stat, []).append(channel_stat)

        # Aggregate all channel indices and stats
        self.ch_map = np.concatenate(channel_map)
        for stat, channel_stat in channel_stats.items():
            aggregated_channel_stat = np.concatenate(channel_stat)
            setattr(self, stat, aggregated_channel_stat)

    def get_channel_mask(self, channels: int, channel: Optional[int] = None) -> np.ndarray:
        """
        Returns a mask for channel stats based on desired channel count and channel
        """
        return (
            self.ch_map[:, 1] == channels
            if channel is None
            else (self.ch_map[:, (1, 2)] == (channels, channel)).all(axis=1)
        )

    def get_image_stats(self) -> dict:
        return {k: getattr(self, k) for k in ImageStats.__annotations__}

    def get_channel_stats(self, channels: int, channel: Optional[int] = None) -> dict:
        return {k: getattr(self, k)[self.get_channel_mask(channels, channel)] for k in ChannelStats.__annotations__}
